/*
	*
	* Copyright 2019 Britanicus <marcusbritanicus@gmail.com>
	*
	* This file is a part of DesQ project ( https://gitlab.com/desq/ )
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include "DesQMenuViews.hpp"

DesQCategoryView::DesQCategoryView( QWidget *parent ) : QListWidget( parent ) {

	setViewMode( QListView::ListMode );
	setIconSize( QSize( 32, 32 ) );
	setGridSize( QSize( 230, 36 ) );
	setFixedWidth( 250 );
	setFrameStyle( QFrame::NoFrame );
	viewport()->setAutoFillBackground( false );
	setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
	setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );

	QStringList categories = {
		"All Applications", "Accessories", "Development", "Education", "Games", "Graphics",
		"Internet", "Multimedia", "Office", "Science & Math", "Settings", "System", "Wine"
	};

	QStringList categoryIcons = {
		"preferences-system-windows", "applications-utilities", "applications-development", "applications-education", "applications-games", "applications-graphics",
		"applications-internet", "applications-multimedia", "applications-office", "applications-science", "preferences-other", "applications-system",
		"wine"
	};

	for( int i = 0; i < categories.count(); i++ ) {
		QListWidgetItem *item = new QListWidgetItem( QIcon::fromTheme( categoryIcons.value( i ) ), categories.at( i ) );
		item->setSizeHint( QSize( 230, 36 ) );
		item->setTextAlignment( Qt::AlignVCenter | Qt::AlignLeft );

		addItem( item );
	}

	connect( this, &QListWidget::itemClicked, this, [=]( QListWidgetItem *item ) {
			emit loadCategory( item->text() );
		}
	);

	setCurrentItem( item( 0 ) );
};

void DesQCategoryView::setCategories( QStringList categories ) {

	if ( not categories.contains( "All Applications" ) )
		categories << "All Applications";

	for( int r = 0; r < 13; r++ ) {
		QListWidgetItem *itm = item( r );
		if ( not categories.contains( itm->text() ) )
			itm->setHidden( true );

		else
			itm->setHidden( false );
	}
};

void DesQCategoryView::mouseReleaseEvent( QMouseEvent *mEvent ) {

	if ( not itemAt( mEvent->pos() ) )
		emit closeMenu();

	QListWidget::mouseReleaseEvent( mEvent );
};

DesQApplicationsView::DesQApplicationsView( QWidget *parent ) : QListWidget( parent ) {

	setViewMode( QListView::IconMode );
	setFlow( QListView::LeftToRight );
	setWrapping( true );
	setWordWrap( false );
	setTextElideMode( Qt::ElideRight );
	setResizeMode( QListView::Adjust );
	setMovement( QListView::Static );
	setIconSize( QSize( 64, 64 ) );
	setGridSize( QSize( 144, 108 ) );
	setFrameStyle( QFrame::NoFrame );
	setUniformItemSizes( true );
	viewport()->setAutoFillBackground( false );

	setItemDelegate( new DesQAppsViewDelegate() );

	parseDesktops();
	loadCategory( "All Applications" );

	QFileSystemWatcher *fsw = new QFileSystemWatcher();
	fsw->addPaths( {
		QDir::home().filePath( ".local/share/applications/" ),
		"/usr/local/share/applications/",
		"/usr/share/applications/" ,
		QDir::home().filePath( ".local/share/games/" ),
		"/usr/local/share/games/",
		"/usr/share/games/"
	} );

	connect( fsw, &QFileSystemWatcher::directoryChanged, [=]( QString ) {
		parseDesktops();
		loadCategory( mCurCategory );
	} );

	connect( this, &QListWidget::itemClicked, this, [=]( QListWidgetItem *item ) {

			QString desktopUrl = item->data( Qt::UserRole + 1 ).toString();

			DesQDesktopFile appFile( desktopUrl );
			appFile.startApplicationWithArgs( {} );

			emit closeMenu();
		}
	);
};

void DesQApplicationsView::loadCategory( QString category ) {

	clear();

	mCurCategory = category;
	if ( category == "All Applications" ) {
		Q_FOREACH( DesQDesktopFile app, allApps ) {

			if ( not mFilter.isEmpty() ) {
				bool matches = false;
				matches |= app.name().contains( mFilter, Qt::CaseInsensitive );
				matches |= app.desktopName().contains( mFilter, Qt::CaseInsensitive );
				matches |= app.command().contains( mFilter, Qt::CaseInsensitive );

				if ( not matches )
					continue;
			}

			QString name = app.name();
			QString icoStr = app.icon();

			/*
				*
				* QIcon bug:
				* If @icoStr starts with /, QIcon::isNull() always returns false,
				* even when we use QIcon::fromTheme( ... )
				*
				* So first check if it begins with /
				*
			*/
			QIcon icon;
			if ( icoStr.startsWith( "/" ) ) {
				if ( DesQUtils::exists( icoStr ) ) {

					icon = QIcon( icoStr );
				}

				else {

					icon = QIcon::fromTheme( "application-x-executable" );
				}
			}

			else if ( QIcon::hasThemeIcon( icoStr ) ) {

				icon = QIcon::fromTheme( icoStr );
			}

			else {

				icon = QIcon::fromTheme( "application-x-executable" );
			}

			QListWidgetItem *item = new QListWidgetItem( icon, name, this );
			item->setData( Qt::UserRole + 1, app.desktopFileUrl() );
			// item->setTextAlignment( Qt::AlignVCenter | Qt::AlignLeft );

			addItem( item );
		}
	}

	else {
		Q_FOREACH( DesQDesktopFile app, categoryAppsMap[ category ] ) {

			if ( not mFilter.isEmpty() ) {
				bool matches = false;
				matches |= app.name().contains( mFilter, Qt::CaseInsensitive );
				matches |= app.desktopName().contains( mFilter, Qt::CaseInsensitive );
				matches |= app.command().contains( mFilter, Qt::CaseInsensitive );

				if ( not matches )
					continue;
			}

			QString name = app.name();
			QString icoStr = app.icon();

			/* Read diatribe starting from line 92 above */
			QIcon icon;
			if ( icoStr.startsWith( "/" ) ) {
				if ( DesQUtils::exists( icoStr ) ) {

					icon = QIcon( icoStr );
				}

				else {

					icon = QIcon::fromTheme( "application-x-executable" );
				}
			}

			else if ( QIcon::hasThemeIcon( icoStr ) ) {

				icon = QIcon::fromTheme( icoStr );
			}

			else {

				icon = QIcon::fromTheme( "application-x-executable" );
			}

			QListWidgetItem *item = new QListWidgetItem( icon, name, this );
			item->setData( Qt::UserRole + 1, app.desktopFileUrl() );
			// item->setTextAlignment( Qt::AlignVCenter | Qt::AlignLeft );

			addItem( item );
		}
	}
};

void DesQApplicationsView::setFilter( QString filter ) {

	mFilter = filter;
	loadCategory( mCurCategory );
};

QStringList DesQApplicationsView::availableCategories() {

	return mAvailableCategories;
};

void DesQApplicationsView::parseDesktops() {

	categoryAppsMap.clear();
	allApps.clear();
	mAvailableCategories.clear();

	/* Apps */
	QStringList appsDirs = {
		QDir::home().filePath( ".local/share/applications/" ),
		"/usr/local/share/applications/",
		"/usr/share/applications/"
	};

	Q_FOREACH( QString appDirLoc, appsDirs ) {
		QDir appDir( appDirLoc );
		appDir.setFilter( QDir::Files );
		appDir.setNameFilters( { "*.desktop" } );

		Q_FOREACH( QString app, appDir.entryList() ) {
			DesQDesktopFile desktopFile( appDir.filePath( app ) );

			QString category = desktopFile.category();
			QString desktopUrl = appDir.filePath( app );

			if ( not desktopFile.isValid() )
				continue;

			if ( desktopFile.type() != DesQDesktopFile::Type::Application )
				continue;

			if ( desktopFile.visible() ) {
				allApps << desktopFile;

				if ( not categoryAppsMap.contains( category ) )
					categoryAppsMap[ category ] = DesQAppsList();

				categoryAppsMap[ category ] << desktopFile;
			}
		}
	}

	/* Games */
	QStringList gamesDirs = {
		QDir::home().filePath( ".local/share/games/" ),
		"/usr/local/share/games/",
		"/usr/share/games/"
	};

	Q_FOREACH( QString gameDirLoc, gamesDirs ) {
		QDir gameDir( gameDirLoc );
		gameDir.setFilter( QDir::Files );
		gameDir.setNameFilters( { "*.desktop" } );

		Q_FOREACH( QString app, gameDir.entryList() ) {
			DesQDesktopFile desktopFile( gameDir.filePath( app ) );

			QString category = desktopFile.category();
			QString desktopUrl = gameDir.filePath( app );

			if ( not desktopFile.isValid() )
				continue;

			if ( desktopFile.type() != DesQDesktopFile::Type::Application )
				continue;

			if ( desktopFile.visible() ) {
				allApps << desktopFile;

				if ( not categoryAppsMap.contains( "Games" ) )
					categoryAppsMap[ "Games" ] = DesQAppsList();

				categoryAppsMap[ "Games" ] << desktopFile;
			}
		}
	}

	/* Wine */
	QDirIterator wineDirIt( QDir::home().filePath( ".local/share/applications/wine/" ), { "*.desktop" }, QDir::Files, QDirIterator::Subdirectories );
	while( wineDirIt.hasNext() ) {

		DesQDesktopFile desktopFile( wineDirIt.next() );

		QString category = desktopFile.category();
		QString desktopUrl = wineDirIt.filePath();

		if ( not desktopFile.isValid() )
			continue;

		if ( desktopFile.type() != DesQDesktopFile::Type::Application )
			continue;

		if ( desktopFile.visible() ) {
			allApps << desktopFile;

			if ( not categoryAppsMap.contains( "Wine" ) )
				categoryAppsMap[ "Wine" ] = DesQAppsList();

			categoryAppsMap[ "Wine" ] << desktopFile;
		}
	}

	mAvailableCategories << categoryAppsMap.keys();

	std::sort( allApps.begin(), allApps.end(), [=]( const DesQDesktopFile first, const DesQDesktopFile second ) {

		return first.name().toLower() < second.name().toLower() ;
	} );

	Q_FOREACH( QString category, mAvailableCategories ) {
		std::sort( categoryAppsMap[ category ].begin(), categoryAppsMap[ category ].end(), [=]( const DesQDesktopFile first, const DesQDesktopFile second ) {

				return first.name().toLower() < second.name().toLower() ;
			} );
	}
};

void DesQApplicationsView::mouseReleaseEvent( QMouseEvent *mEvent ) {

	if ( not itemAt( mEvent->pos() ) )
		emit closeMenu();

	QListWidget::mouseReleaseEvent( mEvent );
};

void DesQApplicationsView::paintEvent( QPaintEvent *pEvent ) {

	QPainter painter( viewport() );
	painter.setPen( Qt::gray );
	painter.drawLine( geometry().topRight(), geometry().bottomRight() );
	painter.end();

	QListWidget::paintEvent( pEvent );
};

void DesQApplicationsView::resizeEvent( QResizeEvent *rEvent ) {

	rEvent->accept();

	QSize vSize = viewport()->size();

	/* Items per row */
	float items = vSize.width() / ( 144 + spacing() );

	/* Minimum width of all items */
	float itemsWidth  = items * 144;

	/* Empty space remaining */
	float empty = vSize.width() - itemsWidth;

	/* Extra space per item */
	float extra = ( empty / items ) - 5;

	/* New grid size */
	setGridSize( QSize( 144 + extra, 108 ) );
};

DesQFavoritesView::DesQFavoritesView( QWidget *parent ) : QListWidget( parent ) {

	setViewMode( QListView::ListMode );
	setIconSize( QSize( 32, 32 ) );
	setGridSize( QSize( 230, 36 ) );
	setFixedWidth( 250 );
	setFrameStyle( QFrame::NoFrame );
	viewport()->setAutoFillBackground( false );

	QSettings favApps( "DesQ", "Menu" );
	Q_FOREACH( QString desktop, favApps.value( "FavoriteApps" ).toStringList() ) {

		DesQDesktopFile appFile( desktop );
		QString name = appFile.name();
		QString icoStr = appFile.icon();
		QIcon icon = QIcon::fromTheme( icoStr );
		if ( icon.isNull() ) {
			if ( DesQUtils::exists( icoStr ) )
				icon = QIcon( icoStr );

			else
				icon = QIcon::fromTheme( "application-x-executable" );
		}

		QListWidgetItem *item = new QListWidgetItem( icon, name );
		item->setData( Qt::UserRole + 1, appFile.desktopFileUrl() );
		item->setSizeHint( QSize( 230, 36 ) );
		item->setTextAlignment( Qt::AlignVCenter | Qt::AlignLeft );

		addItem( item );
	}

	connect( this, &QListWidget::itemClicked, this, [=]( QListWidgetItem *item ) {

			QString desktopUrl = item->data( Qt::UserRole + 1 ).toString();
			DesQDesktopFile appFile( desktopUrl );
			appFile.startApplicationWithArgs( {} );

			emit closeMenu();
		}
	);
};

void DesQFavoritesView::mouseReleaseEvent( QMouseEvent *mEvent ) {

	if ( not itemAt( mEvent->pos() ) )
		emit closeMenu();

	QListWidget::mouseReleaseEvent( mEvent );
};

QSize DesQAppsViewDelegate::sizeHint( const QStyleOptionViewItem &option, const QModelIndex &index ) const {

	QSize size( QStyledItemDelegate::sizeHint( option, index ) );

	/* Icon View */
	if ( option.decorationPosition == QStyleOptionViewItem::Top ) {
        // size.setWidth( static_cast<int>(FileSystemView::gridWidth) - 5 );
		return QSize( 128, 96 );
	}

	/* Tile View */
	else if ( option.decorationPosition == QStyleOptionViewItem::Left ) {
        // size.setWidth( static_cast<int>(FileSystemView::gridWidth) - 5 );
		return QSize( 256, 72 );
	}

	return size;
};
