/*
	*
	* Copyright 2019 Britanicus <marcusbritanicus@gmail.com>
	*
	* This file is a part of DesQ project ( https://gitlab.com/desq/ )
	* Qt5 example of analog clock, and Lenovo K8 Note's clock widget
	* are the inspiration for this. Some parts of the code are taken
	* from Qt5's example
	*
	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*
	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*
	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include <QtWidgets>
#include "VolumeWidget.hpp"

PactlQtDevice::PactlQtDevice( QString name, int id ) {

	mDeviceId = id;
	mName = name;
};

QString PactlQtDevice::name() {

	return mName;
}

int PactlQtDevice::currentVolume() {

	QProcess proc;
	proc.start( "pactl", QStringList() << "list" << "sinks" );
	proc.waitForFinished();

	QStringList data = QString::fromLocal8Bit( proc.readAll() ).split( "\n", QString::SkipEmptyParts );
	Q_FOREACH( QString chunk, data ) {
		if ( not chunk.startsWith( "Sinks: #" + QString::number( mDeviceId ) ) )
			continue;

		if ( chunk.simplified().trimmed().startsWith( "Volume" ) ) {
			QStringList chunks = chunk.simplified().trimmed().split( " ", QString::SkipEmptyParts );
			int left = chunks.value( 4, "-1" ).replace( "%", "" ).toInt();
			int right = chunks.value( 11, "-1" ).replace( "%", "" ).toInt();

			if ( ( left >= 0 ) and ( right >= 0 ) )
                return static_cast<int>( left + right ) / 2;
		}
	}

	return 0;
};

bool PactlQtDevice::isMuted() {

	QProcess proc;
	proc.start( "pactl", QStringList() << "list" << "sinks" );
	proc.waitForFinished();

	QStringList data = QString::fromLocal8Bit( proc.readAll() ).split( "\n", QString::SkipEmptyParts );
	Q_FOREACH( QString chunk, data ) {
		if ( chunk.simplified().trimmed().startsWith( "Mute:" ) ) {
			QStringList chunks = chunk.simplified().trimmed().split( " ", QString::SkipEmptyParts );

			if ( chunks.value( 1 ) == "yes" )
				return true;

			else
				return false;
		}
	}

	/* By default, treat it as muted */
	return true;
};

void PactlQtDevice::setVolume( int volume ) {

	QProcess proc;
	proc.startDetached( "pactl", QStringList() << "set-sink-volume" << "@DEFAULT_SINK@" << QString( "%1%" ).arg( volume ) );
};

void PactlQtDevice::toggleMute() {

	QProcess proc;
	proc.startDetached( "pactl", QStringList() << "set-sink-mute" << "@DEFAULT_SINK@" << QString( "toggle" ) );
};

VolumeWidget::VolumeWidget( QWidget *parent ) : QWidget( parent ) {

	display = new QLabel( this );
	display->setPixmap( QPixmap( ":/speaker.png" ).scaled( 84, 84, Qt::KeepAspectRatio, Qt::SmoothTransformation ) );
	display->setFixedSize( 100, 100 );

	setFixedSize( 200, 120 );

	QProcess proc;
	proc.start( "pactl", QStringList() << "list" << "short" << "sinks" );
	proc.waitForFinished();

	proc.start( "pactl", QStringList() << "list" << "sinks" );
	proc.waitForFinished();

	QStringList devProps = QString::fromLocal8Bit( proc.readAll() ).split( "\n\n", QString::SkipEmptyParts );

	// To get the name of the device
	QStringList props = devProps.value( 0 ).split( "\n", QString::SkipEmptyParts );
	QString name;
	Q_FOREACH( QString prop, props ) {
		if ( prop.trimmed().simplified().startsWith( "device.description" ) ) {
			name = prop.trimmed().simplified().replace( "\"", "" ).split( " = ", QString::SkipEmptyParts ).value( 1 );
			break;
		}
	}

	device = new PactlQtDevice( name, 0 );

	volBar = new QProgressBar();
	volBar->setRange( 0, 100 );
	volBar->setFormat( "" );
	volBar->setOrientation( Qt::Vertical );
	volBar->setValue( device->currentVolume() );

	QHBoxLayout *lyt = new QHBoxLayout();
	lyt->addStretch();
	lyt->addWidget( display, Qt::AlignCenter );
	lyt->addWidget( volBar, Qt::AlignCenter );
	lyt->addStretch();

	setLayout( lyt );

	opacity = new QGraphicsOpacityEffect( this );
	opacity->setOpacity( 0.4 );
	setGraphicsEffect( opacity );
};

void VolumeWidget::wheelEvent( QWheelEvent *wEvent ) {

	int angle = wEvent->angleDelta().y() / 8;

	int newVol = device->currentVolume() + angle;
	newVol = qMin( 100, qMax( 0, newVol ) );

	device->setVolume( newVol );
	volBar->setValue( newVol );

	wEvent->accept();
};

void VolumeWidget::enterEvent( QEvent *event ) {

	opacity->setOpacity( 1.0 );
	event->accept();
};

void VolumeWidget::leaveEvent( QEvent *event ) {

	opacity->setOpacity( 0.4 );
	event->accept();
};
